import logging
from functools import partial
from http import HTTPStatus
from typing import Iterator, List, Optional, Set, Tuple

import requests

from ...utils import (
    APIClient,
    BearerAuth,
    SafeMode,
    build_url,
    fetch_all_pages,
    handle_response,
    retry,
    retry_request,
    safe_mode,
)
from ..abstract import TimeFilter
from .credentials import DatabricksCredentials
from .endpoints import DatabricksEndpointFactory
from .format import DatabricksFormatter, TagMapping
from .lineage import single_column_lineage_links, single_table_lineage_links
from .pagination import DATABRICKS_PAGE_SIZE, DatabricksPagination
from .types import TablesColumns, TimestampedLink
from .utils import hourly_time_filters

logger = logging.getLogger(__name__)

_DATABRICKS_CLIENT_TIMEOUT_S = 90
_MAX_NUMBER_OF_LINEAGE_ERRORS = 1000
_MAX_NUMBER_OF_QUERY_ERRORS = 1000
_RETRY_ATTEMPTS = 3
_RETRY_BASE_MS = 1000
_RETRY_EXCEPTIONS = [
    requests.exceptions.ConnectTimeout,
]
_WORKSPACE_ID_HEADER = "X-Databricks-Org-Id"

safe_lineage_params = SafeMode((BaseException,), _MAX_NUMBER_OF_LINEAGE_ERRORS)
safe_query_params = SafeMode((BaseException,), _MAX_NUMBER_OF_QUERY_ERRORS)


class DatabricksAuth(BearerAuth):
    def __init__(self, credentials: DatabricksCredentials):
        self.token = credentials.token

    def fetch_token(self) -> Optional[str]:
        return self.token


class DatabricksAPIClient(APIClient):
    """Databricks Client"""

    def __init__(
        self,
        credentials: DatabricksCredentials,
        db_allowed: Optional[Set[str]] = None,
        db_blocked: Optional[Set[str]] = None,
    ):
        auth = DatabricksAuth(credentials)
        super().__init__(
            host=credentials.host,
            auth=auth,
            timeout=_DATABRICKS_CLIENT_TIMEOUT_S,
        )
        self._http_path = credentials.http_path
        self._db_allowed = db_allowed
        self._db_blocked = db_blocked

        self.formatter = DatabricksFormatter()

    def _keep_catalog(self, catalog: str) -> bool:
        """
        Helper function to determine if we should keep the Databricks catalog
        which is a CastorDoc database
        """
        if self._db_allowed and catalog not in self._db_allowed:
            return False
        if self._db_blocked and catalog in self._db_blocked:
            return False
        return True

    def databases(self) -> List[dict]:
        content = self._get(DatabricksEndpointFactory.databases())
        _databases = self.formatter.format_database(content.get("catalogs", []))
        return [d for d in _databases if self._keep_catalog(d["database_name"])]

    def _schemas_of_database(self, database: dict) -> List[dict]:
        payload = {"catalog_name": database["database_name"]}
        content = self._get(DatabricksEndpointFactory.schemas(), params=payload)
        schemas = content.get("schemas", [])
        return self.formatter.format_schema(schemas, database)

    def schemas(self, databases: List[dict]) -> List[dict]:
        """
        Get the databricks schemas (also sometimes called databases)
        (which correspond to the schemas in Castor)
        leveraging the unity catalog API
        """
        return [
            schema
            for database in databases
            for schema in self._schemas_of_database(database)
        ]

    def tables_columns_of_schema(
        self,
        schema: dict,
        table_tags: TagMapping,
        column_tags: TagMapping,
    ) -> TablesColumns:
        payload = {
            "catalog_name": schema["database_id"],
            "schema_name": schema["schema_name"],
        }
        response = self._call(
            method="GET",
            endpoint=DatabricksEndpointFactory.tables(),
            params=payload,
        )
        workspace_id = response.headers[_WORKSPACE_ID_HEADER]
        content = handle_response(response)
        host = build_url(self._host, endpoint="")
        return self.formatter.format_table_column(
            raw_tables=content.get("tables", []),
            schema=schema,
            host=host,
            workspace_id=workspace_id,
            table_tags=table_tags,
            column_tags=column_tags,
        )

    @safe_mode(safe_lineage_params, lambda: [])
    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_ATTEMPTS,
        base_ms=_RETRY_BASE_MS,
    )
    @retry_request(
        status_codes=(HTTPStatus.TOO_MANY_REQUESTS,),
        max_retries=_RETRY_ATTEMPTS,
    )
    def get_single_column_lineage(
        self,
        names: Tuple[str, str],
    ) -> List[TimestampedLink]:
        """
        Helper function used in get_lineage_links.
        Call data lineage API and return the content of the result

        eg table_path: broward_prd.bronze.account_adjustments
        FYI: Maximum rate of 10 requests per SECOND
        """
        table_path, column_name = names
        payload = {
            "table_name": table_path,
            "column_name": column_name,
            "include_entity_lineage": True,
        }
        content = self._get(
            DatabricksEndpointFactory.column_lineage(), params=payload
        )
        column_path = f"{table_path}.{column_name}"
        return single_column_lineage_links(column_path, content)

    @safe_mode(safe_lineage_params, lambda: [])
    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_ATTEMPTS,
        base_ms=_RETRY_BASE_MS,
    )
    def get_single_table_lineage(
        self, table_path: str
    ) -> List[TimestampedLink]:
        """
        Helper function used in get_lineage_links.
        Call data lineage API and return the content of the result
        eg table_path: broward_prd.bronze.account_adjustments
        FYI: Maximum rate of 50 requests per SECOND
        """
        payload = {"table_name": table_path, "include_entity_lineage": True}
        content = self._get(
            DatabricksEndpointFactory.table_lineage(), params=payload
        )
        return single_table_lineage_links(table_path, content)

    @safe_mode(safe_query_params, lambda: [])
    @retry(
        exceptions=_RETRY_EXCEPTIONS,
        max_retries=_RETRY_ATTEMPTS,
        base_ms=_RETRY_BASE_MS,
    )
    def _queries(
        self,
        filter_: dict,
    ) -> Iterator[dict]:
        """
        Callback to scroll the queries api
        https://docs.databricks.com/api/workspace/queryhistory/list
        max_results: Limit the number of results returned in one page.
            The default is 100. (both on our side and Databricks')
        """
        payload = {**filter_, "max_results": DATABRICKS_PAGE_SIZE}
        request = partial(
            self._get,
            endpoint=DatabricksEndpointFactory.queries(),
            data=payload,
        )
        queries = fetch_all_pages(request, DatabricksPagination)
        return queries

    def queries(self, time_filter: Optional[TimeFilter] = None) -> List[dict]:
        """get all queries, hour per hour"""
        time_range_filters = hourly_time_filters(time_filter)
        raw_queries = []
        for _filter in time_range_filters:
            logger.info(f"Fetching queries for time filter {_filter}")
            hourly = self._queries(_filter)
            raw_queries.extend(hourly)
        return self.formatter.format_query(raw_queries)

    def users(self) -> List[dict]:
        """
        retrieve user from api
        """
        content = self._get(DatabricksEndpointFactory.users())
        return self.formatter.format_user(content.get("Resources", []))

    def _view_ddl_per_schema(self, schema: dict) -> List[dict]:
        payload = {
            "catalog_name": schema["database_id"],
            "schema_name": schema["schema_name"],
            "omit_columns": True,
        }
        content = self._get(DatabricksEndpointFactory.tables(), params=payload)
        return self.formatter.format_view_ddl(content.get("tables", []), schema)

    def view_ddl(self, schemas: List[dict]) -> List[dict]:
        """retrieve view ddl"""
        view_ddl: List[dict] = []
        for schema in schemas:
            v_to_add = self._view_ddl_per_schema(schema)
            view_ddl.extend(v_to_add)
        return view_ddl



from typing import Dict, List, Set, Tuple, cast

from .types import Link, Ostr, OTimestampedLink, TimestampedLink


class LineageLinks:
    """
    helper class that handles lineage deduplication and filtering
    """

    def __init__(self):
        self.lineage: Dict[Link, Ostr] = dict()

    def add(self, timestamped_link: TimestampedLink) -> None:
        """
        keep the most recent lineage link, adding to `self.lineage`
        """
        parent, child, timestamp = timestamped_link
        link = (parent, child)
        if not self.lineage.get(link):
            self.lineage[link] = timestamp
            return

        if not timestamp:
            return
        # keep most recent link; cast for mypy
        recent = max(cast(str, self.lineage[link]), cast(str, timestamp))
        self.lineage[link] = recent


def _to_table_path(table: dict) -> Ostr:
    if table.get("name"):
        return f"{table['catalog_name']}.{table['schema_name']}.{table['name']}"
    return None


def _to_column_path(column: dict) -> Ostr:
    if column.get("name"):
        return f"{column['catalog_name']}.{column['schema_name']}.{column['table_name']}.{column['name']}"
    return None


def _link(path_from: Ostr, path_to: Ostr, timestamp: Ostr) -> OTimestampedLink:
    """exclude missing path and self-lineage"""
    if (not path_from) or (not path_to):
        return None
    is_self_lineage = path_from.lower() == path_to.lower()
    if is_self_lineage:
        return None
    return path_from, path_to, timestamp


def single_table_lineage_links(
    table_path: str, single_table_lineage: dict
) -> List[TimestampedLink]:
    """
    process databricks lineage API response for a given table
    returns a list of (parent, child, timestamp)

    Note: in `upstreams` or `downstreams` we only care about `tableInfo`,
    we could also have `notebookInfos` or `fileInfo`
    """
    links: List[OTimestampedLink] = []
    # add parent:
    for link in single_table_lineage.get("upstreams", []):
        parent = link.get("tableInfo", {})
        parent_path = _to_table_path(parent)
        timestamp: Ostr = parent.get("lineage_timestamp")
        links.append(_link(parent_path, table_path, timestamp))

    # add children:
    for link in single_table_lineage.get("downstreams", []):
        child = link.get("tableInfo", {})
        child_path = _to_table_path(child)
        timestamp = child.get("lineage_timestamp")
        links.append(_link(table_path, child_path, timestamp))

    return list(filter(None, links))


def single_column_lineage_links(
    column_path: str, single_column_lineage: dict
) -> List[TimestampedLink]:
    """
    process databricks lineage API response for a given table
    returns a list of (parent, child, timestamp)

    Note: in `upstreams` or `downstreams` we only care about `tableInfo`,
    we could also have `notebookInfos` or `fileInfo`
    """
    links: List[OTimestampedLink] = []
    # add parent:
    for link in single_column_lineage.get("upstream_cols", []):
        parent_path = _to_column_path(link)
        timestamp: Ostr = link.get("lineage_timestamp")
        links.append(_link(parent_path, column_path, timestamp))

    # add children:
    for link in single_column_lineage.get("downstream_cols", []):
        child_path = _to_column_path(link)
        timestamp = link.get("lineage_timestamp")
        links.append(_link(column_path, child_path, timestamp))

    return list(filter(None, links))


def paths_for_column_lineage(
    tables: List[dict], columns: List[dict], table_lineage: List[dict]
) -> List[Tuple[str, str]]:
    """
    helper providing a list of candidate columns to look lineage for:
    we only look for column lineage where there is table lineage
    """
    # mapping between table id and its path db.schema.table
    # table["schema_id"] follows the pattern `db.schema`
    mapping = {
        table["id"]: ".".join([table["schema_id"], table["table_name"]])
        for table in tables
    }

    tables_with_lineage: Set[str] = set()
    for t in table_lineage:
        tables_with_lineage.add(t["parent_path"])
        tables_with_lineage.add(t["child_path"])

    paths_to_return: List[Tuple[str, str]] = []
    for column in columns:
        table_path = mapping[column["table_id"]]
        if table_path not in tables_with_lineage:
            continue
        column_ = (table_path, column["column_name"])
        paths_to_return.append(column_)

    return paths_to_return


def deduplicate_lineage(lineages: List[TimestampedLink]) -> dict:
    deduplicated_lineage = LineageLinks()
    for timestamped_link in lineages:
        deduplicated_lineage.add(timestamped_link)
    return deduplicated_lineage.lineage



from .lineage import LineageLinks
from .test_constants import (
    CLOSER_DATE,
    OLDER_DATE,
)


def test_LineageLinks_add():
    links = LineageLinks()
    timestamped_link = ("parent", "child", None)
    expected_key = ("parent", "child")

    links.add(timestamped_link)

    assert expected_key in links.lineage
    assert links.lineage[expected_key] is None

    # we replace None by an actual timestamp
    timestamped_link = ("parent", "child", OLDER_DATE)
    links.add(timestamped_link)
    assert expected_key in links.lineage
    assert links.lineage[expected_key] == OLDER_DATE

    # we update with the more recent timestamp
    timestamped_link = ("parent", "child", CLOSER_DATE)
    links.add(timestamped_link)
    assert expected_key in links.lineage
    assert links.lineage[expected_key] == CLOSER_DATE

    # we keep the more recent timestamp
    timestamped_link = ("parent", "child", OLDER_DATE)
    links.add(timestamped_link)
    assert expected_key in links.lineage
    assert links.lineage[expected_key] == CLOSER_DATE



import csv
from io import StringIO
from typing import Iterator


def usage_liveboard_reader(usage_liveboard_csv: str) -> Iterator[dict]:
    """
    Converts a CSV string into an iterator of dictionaries after
    ignoring the first 6 lines, using the 7th line as the header.
    First 6 lines looks like the following:

        "Data extract produced by Castor on 09/19/2024 06:54"
        "Filters applied on data :"
        "User Action IN [pinboard_embed_view,pinboard_tspublic_no_runtime_filter,pinboard_tspublic_runtime_filter,pinboard_view]"
        "Pinboard NOT IN [mlm - availability pinboard,null]"
        "Timestamp >= 20240820 00:00:00 < 20240919 00:00:00"
        "Timestamp >= 20240919 00:00:00 < 20240920 00:00:00"

    """
    csv_file = StringIO(usage_liveboard_csv)

    for _ in range(7):
        next(csv_file)

    yield from csv.DictReader(csv_file)
